<?php
/*
 * Plugin Name: Cycle Online Payment Gateway
 * Plugin URI:
 * Description: Cycle Online Crypto Payment gateway for WooCommerce.
 * Version: 1.0.0
 * Author: Cyclebit
 * License: MIT
 * WC requires at least: 5.0.0
 * WC tested up to: 5.1.0
 */


if (!defined('ABSPATH')) exit; // exit if accessed directly

add_action('plugins_loaded', 'cycle_payment_gateway_init');


global $cyclebit_db_version;
$cyclebit_db_version = '1.1';

/**
 * Creates table in database for collecting the transactions
 *
 * @return void
 */
function cyclebit_db_install() {
    global $wpdb;
    global $cyclebit_db_version;

    $table_name = $wpdb->prefix . 'cyclegateway_transactions';
    
    $charset_collate = $wpdb->get_charset_collate();
  
    $sql = "CREATE TABLE $table_name (
          `id` INT NOT NULL AUTO_INCREMENT,
          `order_id` INT NOT NULL,
          `transaction_id` VARCHAR(45) NOT NULL,          
          `createdate` DATETIME NOT NULL,
          `statusdate` DATETIME NULL,
          `status` VARCHAR(15) NOT NULL,
          `pending` TINYINT(1) NULL DEFAULT 1,
        PRIMARY KEY (`id`)) $charset_collate;";

    require_once( ABSPATH . 'wp-admin/includes/upgrade.php' );
    dbDelta( $sql );

    add_option( 'cyclebit_db_version', $cyclebit_db_version );
}



function cycle_payment_gateway_update_db_check() {
    global $cyclebit_db_version;
    if ( get_site_option( 'cyclebit_db_version' ) != $cyclebit_db_version ) {
        cyclebit_db_install();
    }
}


function cycle_payment_gateway_init()
{
    cycle_payment_gateway_update_db_check();
    if (!class_exists('WC_Payment_Gateway')) {
        return;
    }

    if (class_exists('WC_Cycle_Gateway')) {
        return;
    }
    if (!class_exists('Firebase\JWT\JWT')) {
        require_once(plugin_dir_path(__FILE__) . "vendor/autoload.php");
    }

    if (!class_exists('CycleGateway\Order')) {
        require_once(plugin_dir_path(__FILE__) . "vendor/CycleGateway/Order.php");
    }
    if (!class_exists('CycleGateway\Client')) {
        require_once(plugin_dir_path(__FILE__) . "vendor/CycleGateway/Client.php");
    }

    /**
     * Add the gateway to WooCommerce
     */
    add_filter('woocommerce_payment_gateways', function($methods) {
        $methods[] = 'WC_Cycle_Gateway';
        return $methods;
    });

    /**
     * Gateway class.
     */
    class WC_Cycle_Gateway extends WC_Payment_Gateway 
    {
        /** @var boolean Whether or not logging is enabled */
        public static $log_enabled = false;

        /** @var WC_Logger Logger instance */
        public static $log = false;

        private $gatewayClient;

        /**
        * Throw error on object clone
        *
        * @access public
        * @return void
        */
        public function __clone() {
            trigger_error('Cloning instances of the class is forbidden.');
        }

        /**
        * Disable unserializing of the class
        *
        * @access public
        * @return void
        */
        public function __wakeup() {
            trigger_error('Unserializing instances of the class is forbidden.');
        }

        public function __construct() {
            $this->order_id = 0;
            $this->id = 'cyclegateway';
            $this->method_title = 'Cycle Online';
            $this->method_description = $this->method_title . ' Payment Gateway';

            // Load the settings
            $this->init_form_fields();
            $this->init_settings();

            $this->icon = apply_filters( 'woocommerce_'. $this->id .'_gateway_icon', plugins_url('assets/images/logo.png', __FILE__) );
            $this->supports = array(
                'products'
            );

            foreach ( $this->settings as $setting_key => $value ) {
                $this->$setting_key = $value;
            }
            
            $this->gatewayClient = new CycleGateway\Client(
                isset($this->agent_email) ? $this->agent_email : "", 
                isset($this->secret_key) ? $this->secret_key : "",                
                isset($this->redirect_uri) ? $this->redirect_uri : "",
                isset($this->server_secret_key) ? $this->server_secret_key : "",
            );

            if (!$this->is_valid_for_use()) {
                $this->enabled = 'no';
            }

            self::$log_enabled = $this->debug === 'yes';

            // Save settings
            if ( is_admin() ) {
                // Save administration options.
                add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );
            }
            add_action('woocommerce_receipt_' . $this->id, array( $this, 'receipt_page' ) );
            add_action('woocommerce_api_' . strtolower(get_class($this)), array( $this, 'check_response' ) ); // Payment listener/API hook
            

            if ( is_admin() ) {
                add_action( 'woocommerce_admin_order_data_after_billing_address', array( $this, 'after_order_details' ), 10, 3 );
                
              
            }

            // Check for SSL
            if ($this->gatewayClient->getUserIp() != '127.0.0.1') {
                add_action( 'admin_notices', array( $this, 'do_ssl_check' ) );
            }

           // add_action('wp_enqueue_scripts', array($this, 'current_gateway_checkout_scripts'));
           // add_action('admin_footer', array($this, 'current_gateway_load_admin_js'));

            if ( isset($_GET) && !empty($_GET) ) { 
                // wc_clear_notices();
                $this->orderReceivedPageCalledByGateway($_GET);
            }
        }

        /**
         * Logging method
         * @param  string $message
         */
        public static function log($message) {
            if (self::$log_enabled) {
                $logger = wc_get_logger();
                $logger->info( $message, array( 'source' => 'cyclegateway' ) );
            }
        }      

        public function after_order_details($order)
         {
            $order_id = $order->get_order_number();
            $transactions = $this->getTransactions($order_id, 0);

            if($transactions ){
                $text = '<p></p><strong>Transactions (Cycle Online):</strong><table>';
                foreach ($transactions as  $transaction) {
                    $text .= '<tr>';
                    $text .= '<td style="border-bottom:1px solid #ccc">'. $transaction->transaction_id.'</td>';
                    $text .= '<td style="border-bottom:1px solid #ccc">'. $transaction->status.'</td>';
                    $text .= '</tr>';
                 }  
                 $text .= '</table>';  

                 echo wp_kses_post($text); 
             } 
         } 
        

        /**
         * Enqueue checkout page styles, scripts.
         */
        public function current_gateway_checkout_scripts() {
            if (is_checkout()) {
                wp_enqueue_style($this->id . '-checkout', plugin_dir_url(__FILE__) . 'assets/css/styles.css', array(), '1.0.0');
            }
        }

        /**
         * Load JS for Gateway woo admin area.
        */
        public function current_gateway_load_admin_js() {
            if ( !isset($_GET['tab']) || $_GET['tab'] != 'checkout' || !isset($_GET['section']) || $_GET['section'] != $this->id ) {
                return;
            }

            wp_register_script($this->id . '-scripts', plugin_dir_url(__FILE__) . 'assets/js/scripts.js', array(), '1.0.0', true);
            wp_enqueue_script($this->id . '-scripts');
        }

        /**
         * Check if store currency is supported
         * @return bool true if currency is in supported currencies list
         */
        public function is_valid_for_use() {
            return true; // Allways true. Gateway decides on further stages.
        }

        /**
         * Admin Panel Options
         */
        public function admin_options() {
            if ($this->is_valid_for_use()) {
                parent::admin_options();
            }
            else {
                ?>
                <div class="inline error">
                    <p><strong><?php _e('Gateway Disabled', 'woocommerce'); ?></strong>: 
                    <?php // _e('CycleGateway does not support your store currency.', 'woocommerce'); ?>
                    </p>
                </div>
                <?php
            }
        }

        /**
         * Initialise Gateway Settings Form Fields
         *
         * @access public
         * @return void
         */
        public function init_form_fields() {
            $this->form_fields = include('form-fields.php');

        }

        /**
         * Validate credit card data
         *
         * @return void
         */
        public function validate_fields() {
            if ($this->has_fields) {

            }
        }

        /**
         * Add notice in admin section if gateway is used without SSL certificate
         */
        public function do_ssl_check() {
            if ( $this->enabled == "yes" && get_option( 'woocommerce_force_ssl_checkout' ) == "no" ) {
                echo wp_kses_post("<div class=\"error\"><p>". sprintf( __( "<strong>%s</strong> is enabled and WooCommerce is not forcing the SSL certificate on your checkout page. Please ensure that you have a valid SSL certificate and that you are <a href=\"%s\">forcing the checkout pages to be secured.</a>" ), $this->method_title, admin_url( 'admin.php?page=wc-settings&tab=checkout' ) ) ."</p></div>");
            }
        }
        

        /**
         * Process fields values of this plugin admin settings.
         *
         */
        public function process_admin_options() {
            parent::process_admin_options();
            
        }

        /**
         * Customize Thank you text.
         *
         */
        // public function change_order_received_text($text) {
        //     $newText = $text;
        //     $newText .= '<div style="display: flex;flex-wrap: wrap;">';
        //     foreach ($this->order_payments_statuses as $status) {
        //         $newText .= '<div style="width: 33%;">' . $status['transaction_id'].'</div>';
        //         $newText .= '<div style="width: 33%;">' . $status['status'].'</div>';
        //         $newText .= '<div style="width: 33%;">' . $status['status']['message']['string'].'</div>';
        //     }
        //     $newText .= '</div>';
        //     //$newText .= '<p>' . sprintf(__('. <br />Your payment is beeing processed via %s Payment Gateway.', 'woocommerce-cyclegateway'), $this->method_title) . '</p>';
        //     return $newText;
        // }


        /**
         * Handling payment and processing the order.
         *
         * @param int $order_id Id of the order
         * @return array
         */
        public function process_payment($order_id) {
            self::log("--- Shop order " . $order_id);
            
            $order = new WC_Order($order_id);
            
            $returnArray = [
                'result' => 'success',
                'redirect' => $order->get_checkout_payment_url(true)
            ]; 

            return $returnArray;
        }

        /**
         * Creates order on Gateway. Sets Shop's order status. Shows additional HTML or redirects to Gateway.
         * /checkout/order-pay/SSS/?key=wc_order_ZZZ
         * 
         * @param $order_id
         */
        function receipt_page( $order_id ) {
            self::log("receipt_page : " . $order_id);

            $order = new WC_Order($order_id);

            $order_data = $order->get_data();
            $customerArray = $this->createCustomerArray($order_data);

            $jsonData = $this->createJsonAsArrayForGateway($order, $customerArray);

            $output = $this->gatewayClient->createPayment($jsonData);
            
            $outputCode = (int)$output['code'];
            
            self::log( print_r($output, true) );
            self::log($order_id . ': Output code: ' . $outputCode);
            

            if ( $outputCode == 200 ) {

                if ( isset($output['body']) ) {
                	$gatewayOutputBody = json_decode($output['body'], true);
                }
                else {
                    self::log("ERR: no output body");
                    return $returnArray;
                }

                $gatewayOrderStatus = $this->gatewayClient->getGatewayOrderStatus($gatewayOutputBody);
                self::log($order_id . ": Selfhosted, POST, 200");

                $this->addTransaction($order_id, $gatewayOutputBody['tranId'], $jsonData);
             
           
                $wooOrderStatus = $this->convertGatewayToWooStatus($gatewayOrderStatus);
                $messageEntity = $this->setWooNotice($wooOrderStatus, $gatewayOrderStatus);

                $order->update_status( $wooOrderStatus, $messageEntity['string'] );
                
                $gatewayProcessPageUrl = $gatewayOutputBody['url'];

                if (!$gatewayProcessPageUrl) {
                    self::log("receipt_page, BAD Location : " . $order_id);
                    wc_add_notice( __('Location value not defined. ', 'woocommerce-cyclegateway') . __('Order ID: ', 'woocommerce-cyclegateway') . $order_id, 'error' );
                    $gatewayProcessPageUrl = $this->get_return_url($order);
                }

                $gatewayProcessPageUrl = esc_url( $gatewayProcessPageUrl );

                $order->update_status('pending');
                $order->add_order_note( __('Transaction created - '  . $gatewayOutputBody['tranId'] .'. Waiting for payment.', 'woocommerce-cyclegateway') );

                wp_redirect( $gatewayProcessPageUrl );
                exit;
                
            }   

            //if (function_exists('wc_reduce_stock_levels')) {
            //    wc_reduce_stock_levels($order_id);
            //}

        }


        
        /**
         * Callbakc from gateway
         *  
	        {
	            "TransactionId": "e3cc51a0-4542-4433-af5e-28f2dea99e7e",
	            "State": "completed",
	            "Success": true
	        }
        
            created — payment was created, the payer has not yet selected the currency for payment;
            pending — payer chose the currency to pay the invoice, but incoming transaction in the blockchain has not yet been detected;
            paid — incoming transaction in blockchain was found, but did not reach the required amount of confirmations;
            completed — payment is successfully completed, you can release good(s)/service(s) to the payer, the receipt has been sent to the payer's email address specified in the KYC form;
            failed — payment failed for some reason (verification error, funds were not sent by the payer, etc.).
        */
        public function check_response() {           

            $data = json_decode(file_get_contents('php://input'), true);

            $data_san = [];
            foreach ($data as $key => $value) {
               $data_san[strtolower($key)] = is_string($value)?strtolower($value):$value;
            }

            if(isset($data_san['transactionid']) ){
            	$gateway_response = $this->gatewayClient->getPaymentStatus($data_san['transactionid']);
            	$transaction_status = json_decode($gateway_response['body'], true);

            	if($transaction_status['success'] === true ) {            		
	            	$order_id = $this->getOrderByTransaction($data_san['transactionid']);                   
		            $order = new WC_Order($order_id); 

	                switch ( $transaction_status['state'] /*$data_san['state'] */ ) {
	                	case 'failed':
	                		$this->updateTransaction($data_san['transactionid'], $transaction_status['state'], 0);
	                		if($order) $order->add_order_note( __('Notification was received from the payment gateway.<br>Transaction - '  . $data_san['transactionid'] .'.<br>Status: '. $transaction_status['state'], 'woocommerce-cyclegateway') );
	                		break;
	                	case 'completed':                		
	                		$this->updateTransaction($data_san['transactionid'], $transaction_status['state'], 0); 
		                    if($order){
		                    	$order->add_order_note( __('Notification was received from the payment gateway.<br>Order paid, transaction - '  . $data_san['transactionid'] .'.', 'woocommerce-cyclegateway') );
		                    	$order->payment_complete($data_san['transactionid']);
		                    }
		                   	                    
		                    $this->stopCheck($order_id);
	                		break;
                        case 'paid':
                            if($this->confirm_on_paid == 'yes' && $order) {
                                $order->payment_complete($data_san['transactionid']);
                                $this->stopCheck($order_id);
                            }
	                	case 'pending':
	                	case 'created':               	
	                	default:
	                		if($order) $order->add_order_note( __('Notification was received from the payment gateway.<br>Transaction - '  . $data_san['transactionid'] .'.<br>Status: '. $transaction_status['state'], 'woocommerce-cyclegateway') );
	                		$this->updateTransaction($data_san['transactionid'], $transaction_status['state'], 1); 
	                		break;
	                }    
	            }

            }
            return true;
        }

        /**
         * "Order received" page called from Gateway. Change status of Shop's order. Prepare notice.
         *
         */
        private function orderReceivedPageCalledByGateway($data) {
            global $wpdb;
            
            // - parse url varibles, get order_id (gateway order id)
            if ( !isset($data['shop_order_id']) && !isset($data['order-received'])) {
                return false;
            }

            if ( !isset($data['key']) ) {
                return false;
            }

        }

        private function getTransactions($order_id, $pendingOnly = 1)
        {
            global $wpdb;
            $table = $wpdb->prefix.'cyclegateway_transactions';
            $sql = "
                SELECT transaction_id, status, pending  
                FROM $table
                WHERE order_id = $order_id 
                ";
            if($pendingOnly){
                $sql .= " AND pending = 1";
            }
                
            return $wpdb->get_results($sql);
            
        }

        private function getOrderByTransaction($transaction)
        {
            global $wpdb;
            $table = $wpdb->prefix.'cyclegateway_transactions';
            $sql = "
                SELECT order_id 
                FROM $table
                WHERE transaction_id = '$transaction'
                ";
            
                
            return $wpdb->get_var($sql);
            
        }

        private function addTransaction($order_id, $transaction_id, $data = '')
        {
            global $wpdb;
            $table = $wpdb->prefix.'cyclegateway_transactions';
            $data = array('order_id' => $order_id, 'transaction_id' => $transaction_id, 'pending' => 1, 'createdate' => current_time('mysql', 1), 'statusdate' => current_time('mysql', 1), 'status' => 'created');
            
            $wpdb->insert($table,$data);
        }

        private function updateTransaction($transaction_id, $status, $pending = 1)
        {
            global $wpdb;
            $table = $wpdb->prefix.'cyclegateway_transactions';
            $data = array('pending' => $pending, 'statusdate' => current_time('mysql', 1), 'status' => $status);
            $where = array('transaction_id' => $transaction_id);
            $wpdb->update($table,$data, $where);
        }


        private function stopCheck($order_id)
        {
            global $wpdb;
            $table = $wpdb->prefix.'cyclegateway_transactions';
            $data = array('pending' => 0);
            $where = array('order_id' => $order_id);
            $wpdb->update($table,$data, $where);
        }


        /**
         * Change metadata/data of Shop's Order.
         *
         */
        private function changeShopOrder($orderData, $gatewayOrderId = "", $shop_order_id = 0) {
            if ( !isset($orderData['state']) ) {
                return 0;
            }
            
            if (!$shop_order_id) {
                return 0;
            }

            if (!$gatewayOrderId) {
                $gatewayOrderId = $orderData['tranId'];
            }
            $gatewayOrderUpdated = "";
            
            $gatewayOrderStatus = sanitize_text_field( $orderData['state'] );
           
            $order = new WC_Order($shop_order_id);

            if ( !$order ) {
                self::log("ERR, No order by ID: " . $shop_order_id);
                return 0;
            }

            self::log("upd Shop: " . $shop_order_id . " | Gw: " . $gatewayOrderId);

            if ( 
                $gatewayOrderStatus == 'completed' && 
                function_exists('wc_reduce_stock_levels')
            ) {
                wc_reduce_stock_levels($shop_order_id);
            }

            return $shop_order_id;
        }

        /**
         * Create customer data array for request to Gateway.
         * 
         */

        // "fullname": "John Johnson" – payer’s name (string);
        // "dob": "2000-10-02" – payer’s date of birth in the YYYY-MM-DD format (string);
        // "fullname": "John Johnson" – payer’s name (string);
        // "email": "payer@gmail.com" – payer’s email address (string);
        // "address": "Test Address 11" – payer’s residential address (string);
        // "city": "Miami" – payer’s city of residence (string);
        // "state": "Florida" – payer’s state/province of residence (string);
        // "zip": "33140" – payer’s zip/postal code (string).
        private function createCustomerArray($order_data) {
            if (!$order_data) {
                return array();
            }
            return array(
                'address'   => trim( $order_data['billing']['address_1'] . ' ' . $order_data['billing']['address_2'] ),
                'city'      => $order_data['billing']['city'],
                'country'   => $order_data['billing']['country'],
                'fullname'  => $order_data['billing']['first_name'] . ' ' . $order_data['billing']['last_name'],
                'state'     => $order_data['billing']['state'], 
                'zip'       => $order_data['billing']['postcode'],
                'email'     => $order_data['billing']['email']
            );


        }

        /**
         * Prepare gateway order object for request to Gateway.
         * 
         */
        private function createJsonAsArrayForGateway($order, $customerArray) {
            $orderId = $order->get_order_number();
            
            $gatewayOrder = new CycleGateway\Order();
            $gatewayOrder->orderId = "WooCommerce order #" . $orderId . " from " . get_bloginfo( 'name' );
            $gatewayOrder->amount = $order->get_total();
            $gatewayOrder->callback_url =  add_query_arg( 'wc-api', 'WC_Cycle_Gateway', home_url( '/' ) );
           
            $gatewayOrder->customerArray = $customerArray;

            if(!$this->redirect_uri){
                $returnFromGatewayPageUrl = $this->get_return_url($order); // Go here from Gateway after payment. Originally "Thank you. Your order has been received." page

                $returnFromGatewayPageUrl = add_query_arg('shop_order_id', $orderId, $returnFromGatewayPageUrl);

                $gatewayOrder->redirect_url = esc_url( $returnFromGatewayPageUrl );
            }else{
                $gatewayOrder->redirect_url = $this->redirect_uri;
            }
            
            // var_dump($gatewayOrder);
            // die();
            return $this->gatewayClient->createOrderJson($gatewayOrder);
        }

        /** 
         * Set Woo notice according to Woo Order status. Visual representation, sets color of message.
         *
         * @access private
         * @return array
         */

        /* 
            null — transaction id not found;
            created — payment was created, the payer has not yet selected the currency for payment;
            pending — payer chose the currency to pay the invoice, but incoming transaction in the blockchain has not yet been detected;
            paid — incoming transaction in blockchain was found, but did not reach the required amount of confirmations;
            completed — payment is successfully completed, you can release good(s)/service(s) to the payer, the receipt has been sent to the payer's email address specified in the KYC form;
            failed — payment failed for some reason (verification error, funds were not sent by the payer, etc.).
        */
        private function setWooNotice($wooOrderStatus = 'failed', $gatewayOrderStatus = "") {
            $messageEntities = [
                'null' => [
                    'type' => 'error',
                    'string' => __('Transaction ID not found', 'woocommerce-cyclegateway')
                ],
                'created' => [
                    'type' => 'notice',
                    'string' => __('Payment was created', 'woocommerce-cyclegateway')
                ],
                'pending' => [
                    'type' => 'notice',
                    'string' => __('Payer chose the currency to pay the invoice, but incoming transaction in the blockchain has not yet been detected', 'woocommerce-cyclegateway') 
                ],
                'paid' => [
                    'type' => 'notice',
                    'string' => __('Incoming transaction in blockchain was found, but did not reach the required amount of confirmations', 'woocommerce-cyclegateway')  
                ],
                'completed' => [
                    'type' => 'success',
                    'string' => sprintf(__('Payment is successfully completed via %s.', 'woocommerce-cyclegateway'), $this->method_title)
                ], 
                'failed' => [
                    'type' => 'error',
                    'string' => ucfirst($gatewayOrderStatus) . '.' 
                ]
            ];

            if ( !isset($messageEntities[$wooOrderStatus]) ) {
                return [
                    'type' => 'notice',
                    'string' => ucfirst($gatewayOrderStatus) . '.' 
                ];
            }
            else {
                return $messageEntities[$wooOrderStatus];
            }
        }

        /**
         * Get Woo order status acording to gateway status.
         *
         */
         /* 
            null — transaction id not found;
            created — payment was created, the payer has not yet selected the currency for payment;
            pending — payer chose the currency to pay the invoice, but incoming transaction in the blockchain has not yet been detected;
            paid — incoming transaction in blockchain was found, but did not reach the required amount of confirmations;
            completed — payment is successfully completed, you can release good(s)/service(s) to the payer, the receipt has been sent to the payer's email address specified in the KYC form;
            failed — payment failed for some reason (verification error, funds were not sent by the payer, etc.).
        */
        private function convertGatewayToWooStatus($gatewayOrderStatus) {
            if ( $gatewayOrderStatus == 'completed' ) {
                return 'processing';
            }
            elseif ( $gatewayOrderStatus == 'paid' ) {
                return 'pending';
            }
            elseif ( $gatewayOrderStatus == 'pending' ) {
                return 'pending';
            }
            elseif ( $gatewayOrderStatus == 'created' ) {
                return 'pending';
            }
            elseif ( $gatewayOrderStatus == 'failed' || $gatewayOrderStatus == 'null') {
                return 'failed';
            }
            else {
                return 'pending';
            }
        }

        /**
         * If Order has only virtual or/and downloadable products, then change Woo Status.
         * 
         */
        private function changeWooStatusIfDigital($order, $previouslyDefinedOrderStatus) {
            $items = $order->get_items();
            
            foreach ($items as $item) {
                if ( '0' != $item['variation_id'] ) {
                    $product = new WC_Product( $item['variation_id'] );
                }
                else {
                    $product = new WC_Product( $item['product_id'] );
                }

                if ( !$product->is_virtual() && !$product->is_downloadable() ) {
                    return $previouslyDefinedOrderStatus;
                }
            }

            if ($previouslyDefinedOrderStatus == 'processing') {
                return 'completed';
            }

            return $previouslyDefinedOrderStatus;
        } 

    }
}



